use axum::{
    extract::{Path, State},
    routing::{get, post},
    Json, Router,
};
use http::StatusCode;
use serde::{Deserialize, Serialize};
use sqlx::{query, PgPool};
use ulid::Ulid;
use url::Url;
use uuid::Uuid;

use crate::error::AppError;

pub fn router(db: PgPool) -> Router {
    Router::new()
        .route("/", post(create_link))
        .route(
            "/:id",
            get(get_link_info).put(update_link).delete(delete_link),
        )
        .with_state(db)
}

#[derive(Serialize)]
struct Link {
    id: Ulid,
    slug: String,
    destination: String,
}

async fn get_link_info(
    State(db): State<PgPool>,
    Path(id): Path<Ulid>,
) -> Result<Json<Link>, AppError> {
    let link = query!(
        "SELECT id, slug, destination FROM link WHERE id = $1",
        Uuid::from(id),
    )
    .fetch_one(&db)
    .await?;

    Ok(Json(Link {
        id: Ulid::from(link.id),
        slug: link.slug,
        destination: link.destination,
    }))
}

#[derive(Deserialize)]
struct CreateLinkRequestBody {
    slug: String,
    destination: Url,
}

async fn create_link(
    State(db): State<PgPool>,
    Json(CreateLinkRequestBody { slug, destination }): Json<CreateLinkRequestBody>,
) -> Result<Json<Link>, AppError> {
    let id = Ulid::new();

    let result = query!(
        "INSERT INTO link (id, slug, destination) VALUES ($1, $2, $3) ON CONFLICT DO NOTHING",
        Uuid::from(id),
        slug,
        destination.to_string(),
    )
    .execute(&db)
    .await?;

    match result.rows_affected() {
        1 => Ok(Json(Link {
            id,
            slug,
            destination: destination.to_string(),
        })),
        0 => Err(AppError::LinkExists(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

#[derive(Deserialize)]
struct UpdateLinkRequestBody {
    destination: Url,
}

async fn update_link(
    State(db): State<PgPool>,
    Path(id): Path<Ulid>,
    Json(UpdateLinkRequestBody { destination }): Json<UpdateLinkRequestBody>,
) -> Result<StatusCode, AppError> {
    let result = query!(
        "UPDATE link SET destination = $2 WHERE id = $1",
        Uuid::from(id),
        destination.to_string(),
    )
    .execute(&db)
    .await?;

    match result.rows_affected() {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::LinkNotFound(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}

async fn delete_link(
    State(db): State<PgPool>,
    Path(id): Path<Ulid>,
) -> Result<StatusCode, AppError> {
    let result = query!("DELETE FROM link WHERE id = $1", Uuid::from(id))
        .execute(&db)
        .await?;

    match result.rows_affected() {
        1 => Ok(StatusCode::NO_CONTENT),
        0 => Err(AppError::LinkNotFound(id)),
        rows => Err(AppError::ImpossibleAffectedRows(rows)),
    }
}
